/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                     University of Edinburgh, UK                       */
/*                      Copyright (c) 1995,1996                          */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission to use, copy, modify, distribute this software and its    */
/*  documentation for research, educational and individual use only, is  */
/*  hereby granted without fee, subject to the following conditions:     */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*  This software may not be used for commercial purposes without        */
/*  specific prior written permission from the authors.                  */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*                       Author :  Paul Taylor and Simon King            */
/*                       Date   :  June 1995                             */
/*-----------------------------------------------------------------------*/
/*                  Stream class auxiliary routines                      */
/*                                                                       */
/*=======================================================================*/

#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include "EST_types.h"
#include "EST_FMatrix.h"
#include "ling_class/EST_Stream.h"
#include "EST_Token.h"
#include "EST_string_aux.h"
#include "ling_class/EST_stream_aux.h"
#include "ling_class/EST_stream_compare.h"
#include "EST_io_aux.h"


int close_enough(EST_Stream_Item &a, EST_Stream_Item &b)
{
    return ((b.start() < a.end()) && (a.start() < b.end()));
}

void function_match(EST_II_KVL &u, EST_Stream &a, EST_Stream &b)
{
    (void)u;
    (void)a;
    (void)b;
#if 0
    EST_Stream_Item *a_ptr;
    EST_TBI *i_ptr;
    int i;

    for (i = 0, a_ptr = a.head(); a_ptr != 0; a_ptr = next(a_ptr), ++i)
    {
	if (a_ptr->f("pos")==1)
	{
	    u.add_item(a_ptr->addr(), -1);
	    for (i_ptr = a_ptr->link(b.stream_name())->head(); i_ptr != 0; 
		 i_ptr = next(i_ptr))
		u.change_val(a_ptr->addr(), a_ptr->link(b.stream_name())->item(i_ptr));
	}
    }
#endif
}

void relation_match(EST_Stream &a, EST_Stream &b)
{
    EST_Stream_Item *a_ptr, *b_ptr;

    for (a_ptr = a.head(); a_ptr != 0; a_ptr = next(a_ptr))
	if (a_ptr->f("pos")==1)
	    for (b_ptr = b.head(); b_ptr != 0; b_ptr = next(b_ptr))
		{
		    if ((b_ptr->f("pos")==1)
			&&(close_enough(*a_ptr, *b_ptr)))
		    {
//			cout << "linked\n";
#if 0
			link(*a_ptr, *b_ptr);
#endif
		    }
		}

//		if ((b.pos(b_ptr->name()))
//		    &&(close_enough(*a_ptr, *b_ptr)))
//		    link(*a_ptr, *b_ptr);
}

EST_Stream_Item *nthpos(EST_Stream &a, int n)
{
    EST_Stream_Item *a_ptr;
    int i = 0;
    for (a_ptr = a.head(); a_ptr != 0; a_ptr = next(a_ptr))
	if (a_ptr->f("pos") == 1)
	{
	    if (n == i)
		return a_ptr;
	    ++i;
	}
    return 0;
}

// measures amount of total overlap between segments
float label_distance1(EST_Stream_Item &ref, EST_Stream_Item &test)
{
    float start, end;

    start = fabs(ref.start() - test.start());
    end = fabs(ref.end() - test.end());

    return (start + end) / ref.dur();
}

// Only penalises a test segment that extends beyond the boundaries of 
// a ref segment
float label_distance2(EST_Stream_Item &ref, EST_Stream_Item &test)
{
    float start, end;

    start = (test.start() < ref.start()) ? ref.start() - test.start() : 0;
    end = (ref.end() < test.end()) ? test.end() - ref.end() : 0;

    return (start + end) / ref.dur();
}

int major_matrix_insertions(EST_FMatrix &m, EST_Stream &ref_lab)
{
    int i, j;
    int n = 0;
    EST_Stream_Item *s;

    for (i = 0; i < m.num_rows(); ++i)
    {
	s = nthpos(ref_lab, i);
//	cout << s->name() << ": f:" << s->f("minor")<< endl;
	if (s->f("minor") == 1)
	    ++n;
	else
	    for (j = 0; j < m.num_columns(); ++j)
		if (m(i, j) > -1.0)
		    ++n;
    }
    return (m.num_rows() - n);
}

int major_matrix_deletions(EST_FMatrix &m, EST_Stream &ref_lab)
{
    int i, j;
    int n = 0;
    EST_Stream_Item *s;

    for (j = 0; j < m.num_columns(); ++j)
    {
	s = nthpos(ref_lab, j);
//	cout << s->name() << ": f:" << s->f("minor")<< endl;
	if (s->f("minor") == 1)
	    ++n;
	else
	    for (i = 0; i < m.num_rows(); ++i)
		if (m(i, j) > -1.0)
		    ++n;
    }

    return (m.num_columns() - n);
}

void threshold_labels(EST_Stream &reflab, float t)
{
    (void)reflab;
    (void)t;
#if 0
    EST_Stream_Item *r_ptr;
    float score=0.0;
    int a;

    for (r_ptr = reflab.head(); r_ptr != 0; r_ptr = next(r_ptr))
	if (r_ptr->f("pos")==1)
	{
	    // REORG - temp comment
//	    score = atof(r_ptr->fields());
	    cout << "score is" << score << endl;

	    a = r_ptr->rlink("blank").first();

	    cout << "score is" << score << " address: " << a << endl;
	    if (score > t)
		cout << "delete\n";
	}
#endif
}

/* Check through relations of each ref label, and make there aren't
multiple ref labels related to the same test label. If this is
discoved, compare scores of all the relevant labels and delete the
relations of all but the lowest.

At this stage, each ref label should have one and only one relation to the
test labels.
*/

void multiple_labels(EST_Stream &reflab)
{
    (void)reflab;
#if 0
    EST_Stream_Item *r_ptr, *s_ptr, *t_ptr;;
    EST_TBI *p;
    int a, pos, i;
    EST_TList<int> la;
    float *score;
    score = new float[reflab.length()];

    for (r_ptr = reflab.head(); r_ptr != 0; r_ptr = next(r_ptr))
	if (r_ptr->f("pos")==1)
	{
	    la.clear(); // clear list and add address of current ref label
	    la.append(r_ptr->addr());
	    a = r_ptr->rlink("test").first();
	    cout << a << endl; 

	    // check remainer of ref labels and add any that have same
	    // relations address as r_ptr.
	    for (s_ptr = next(r_ptr); s_ptr != 0; s_ptr = next(s_ptr))
		if (s_ptr->f("pos")==1)
		    if (s_ptr->rlink("test").first() == a)
			la.append(s_ptr->addr());

	    cout << "la: " << la;
	    if (la.length() > 1) // true if the are multiple relations
	    {
		// find scores of all relevant labels
		for (i = 0, p = la.head(); p!= 0; p = next(p), ++i)
		{
		    t_ptr = reflab.item(la.item(p));

		    // REORG - temp comment
//		    score[i] = atof(reflab.item(la.item(p))->fields());
		}
		pos  = lowest_pos(score, i); // find position of lowest score

		cout << "best is " << pos << endl;
		// delete relations of all but lowest score
		for (i = 0, p = la.head(); p!= 0; p = next(p), ++i)
		    if (i != pos)
		    {
			t_ptr = reflab.item(la.item(p));
			t_ptr->rlink("test").clear();
		    }
	    }
	}
#endif

}

/* Compare 2 sets of labels by matrix method. This involves making a M
(number of test labels) x N (number of ref labels) matrix, and
calculating the distance from each label in the test set to each label
in the reference set. The lowest score for each reference is then
recorded. A test is carried out to make sure that no two reference
labels point to the same test label. Then any ref label above a
certain distance is classified as incorrect. The numbers of insertions
and delations are then calculated.  */

EST_FMatrix matrix_compare(EST_Stream &reflab, EST_Stream &testlab, int method,
		       float t, int v)
{
    int i, j, pos;
    int num_ref, num_test;
    EST_Stream_Item *r_ptr, *t_ptr;
    EST_String fns;
    (void)v;

    num_ref = num_test = 0;

    // calculate size of matrix, based on *significant* labels
    for (r_ptr = testlab.head(); r_ptr != 0; r_ptr = next(r_ptr))
	if (r_ptr->f("pos")==1)
	    ++num_test;

    for (r_ptr = reflab.head(); r_ptr != 0; r_ptr = next(r_ptr))
	if (r_ptr->f("pos")==1)
	    ++num_ref;

    EST_FMatrix m(num_test, num_ref);

    if ((m.num_rows() == 0) || (m.num_columns() == 0))
	return m; // nothing to analyse, hence empty matrix 

    // fill matrix values by comparing each test with each reference
    // reference is columns, test is rows

    for (i = 0, t_ptr = testlab.head(); t_ptr != 0; t_ptr = next(t_ptr))
	if (t_ptr->f("pos")==1)
	{
	    for (j = 0, r_ptr = reflab.head(); r_ptr != 0; r_ptr = next(r_ptr))
		if (r_ptr->f("pos")==1)
		{
		    if (method == 1)
			m(i, j) = label_distance1(*r_ptr, *t_ptr);
		    else if (method == 2)
			m(i, j) = label_distance2(*r_ptr, *t_ptr);
		    else
			cerr << "Unknown comparision method" << method << endl;
		    ++j;
		}
	    ++i;
	}

//    cout << "orig M\n";
//    print_matrix_scores(reflab, testlab, m);
    minimise_matrix_by_column(m);
    minimise_matrix_by_row(m);
    matrix_ceiling(m, t);

    // for each ref label, find closest matching test label.
    for (j = 0, r_ptr = reflab.head(); r_ptr != 0; r_ptr = next(r_ptr))
    {
	if (r_ptr->f("pos")==1)
	{
	    pos = lowest_pos(m, j);
	    // REORG - temp comment
//	    r_ptr->set_field_names(r_ptr->fields() +ftoString(m(pos, j)));
	    ++j;
	}
    }
    return m;
}

void multiple_matrix_compare(EST_StreamList &rlist, EST_StreamList
			     &tlist, EST_FMatrix &m, EST_String rpos,
			     EST_String tpos, int method, float t, int v)
{
    EST_TBI *pr, *pt;
    EST_String filename;
    EST_Stream reflab, testlab;
    EST_StrList rposlist, tposlist, rminorlist, tminorlist;
    float ra, rc, mra, mrc;

    StringtoStrList(rpos, rposlist);
    StringtoStrList(tpos, tposlist);
    StringtoStrList("m l mrb mfb lrb lfb", rminorlist);
    StringtoStrList("m l mrb mfb lrb lfb", tminorlist);

    int tot, del, ins, ltot, ldel, lins, lmdel, mdel, lmins, mins;
    tot = del = ins = mdel = mins = 0;

    for (pt = tlist.head(); pt; pt = next(pt))
    {
	pr = StreamList_ptr_extract(rlist, tlist(pt).name(), TRUE);
	if (pr != 0)
	{
	    reflab = rlist(pr);
	    testlab = tlist(pt);

	    convert_to_broad(reflab, rposlist);
	    convert_to_broad(testlab, tposlist);
	    convert_to_broad(reflab, rminorlist, "minor");
	    convert_to_broad(testlab, tminorlist, "minor");

//	    cout << "ref\n" << reflab;
//	    cout << "test\n" << testlab;

//	    cout << "features\n";
//	    print_stream_features(reflab);
	    
	    m = matrix_compare(reflab, testlab, method, t, v);
	    
	    ltot = m.num_columns();
	    ldel = matrix_deletions(m);
	    lmdel = major_matrix_deletions(m, reflab);
	    lins = matrix_insertions(m);
	    lmins = major_matrix_insertions(m, testlab);

	    print_results(reflab, testlab, m, ltot, ldel, lins, v);
//	    cout << "Major Deletions: " << lmdel << endl << endl;;

	    tot += ltot;
	    del += ldel;
	    mdel += lmdel;
	    ins += lins;
	    mins += lmins;
	}
    }
    
    rc = float(tot - del)/(float)tot * 100.0;
    ra = float(tot - del -ins)/(float)tot *100.0;
    mrc = float(tot - mdel)/(float)tot * 100.0;
    mra = float(tot - mdel - mins)/(float)tot *100.0;
    
    if (v)
    {
	cout << "Total " << tot << " del: " << del << " ins: " << ins << endl;
	cout << "Total " << tot << " major del " << mdel << " major ins" << mins << endl;
    }
    cout << "Correct " << rc << "%    Accuracy " << ra << "%" << endl;
    cout << "Major Correct " << mrc << "%    Accuracy " << mra << "%" << endl;
}

void error_location(EST_Stream &e, EST_FMatrix &m, int ref)
{
    int i;
    EST_Stream_Item *s;

    // reference
    if (ref)
    {
	for (i = 0, s = e.head(); s; s = next(s))
	    if ((int)s->f("pos"))
	    {
		if (column_hit(m, i) >= 0)
		    s->f.set("hit", 1);
		else
		    s->f.set("hit", 0);
		++i;
	    }
    }
    else
	for (i = 0, s = e.head(); s; s = next(s))
	    if ((int)s->f("pos"))
	    {
		if (row_hit(m, i) >= 0)
		    s->f.set("hit", 1);
		else
		    s->f.set("hit", 0);
		++i;
	    }
}

void reassign_links(EST_Stream &a, EST_Stream &b, EST_II_KVL &ua, EST_II_KVL &ub)
{

  (void)a;
  (void)b;
  (void)ua;
  (void)ub;
#if 0    
    EST_Stream_Item *a_ptr, *b_ptr;
    for (a_ptr = a.head(); a_ptr != 0; a_ptr = next(a_ptr))
    {
	a_ptr->link(b.stream_name())->clear();
	if ((a_ptr->f("pos")==1) && (ua.val(a_ptr->addr()) != -1))
	    a_ptr->make_link(b.stream_name(), ua.val(a_ptr->addr()));
    }
    for (b_ptr = b.head(); b_ptr != 0; b_ptr = next(b_ptr))
    {
	b_ptr->link(a.stream_name())->clear();
	if ((b_ptr->f("pos")==1) && (ub.val(b_ptr->addr()) != -1))
	    b_ptr->make_link(a.stream_name(), ub.val(b_ptr->addr()));
    }
#endif
}

void reassign_links(EST_Stream &a, EST_II_KVL &u, EST_String stream_type)
{
    (void)a;
    (void)u;
    (void)stream_type;
#if 0    
    EST_Stream_Item *a_ptr;

    for (a_ptr = a.head(); a_ptr != 0; a_ptr = next(a_ptr))
    {
	a_ptr->link(stream_type)->clear();
	if ((a_ptr->f("pos")==1) && (u.val(a_ptr->addr()) != -1))
	    a_ptr->make_link(stream_type, u.val(a_ptr->addr()));
    }
#endif
}


int commutate(EST_Stream_Item *a_ptr, EST_II_KVL &f1, EST_II_KVL &f2,
	      EST_II_KVL &lref, EST_II_KVL &ltest) 

{ 
    int b, c, d, v;
    (void)lref;			// unused parameter
    
    v = a_ptr->addr();
    b = f2.val(v);
    c = (b == -1) ? -1: ltest.val(b);
    d = (c == -1) ? -1: f1.val(c);
    
    return d;
    
    //	    c = ltest.val(f2.val(v));
    //	    d = f1.val(ltest.val(f2.val(v)));
}

//    REF		TEST
//		f1
//	S ----------->  S
//	^		^
//   lr |		| lt
//	|	f2	|
//	E ----------->  E
//
//	For a given element in E(ref), "e", if
//	lr(e) == f1(lt(f2(e)))
// 	Then ei has been recognised fully.

void print_matrix_scores(EST_Stream &ref, EST_Stream &test, EST_FMatrix &a)
{
    int i, j;
    EST_Stream_Item *r_ptr, *t_ptr;
    
    cout << "      ";
    for (r_ptr = ref.head(); r_ptr != 0; r_ptr = next(r_ptr))
    {	
	if (r_ptr->f("pos")==1)
	{
	    //	cout.width(5);
	    //	cout.setf(ios::right);
	    cout << r_ptr->name() << " ";
	    cout.width(6);
	    cout.setf(ios::right);
	    cout<< r_ptr->end() << " ";
	}
    }
    cout << endl;
    
    for (t_ptr = test.head(), i = 0; i < a.num_rows(); t_ptr = next(t_ptr))
    {
	if (t_ptr->f("pos")==1)
	{
	    cout << t_ptr->name() << " ";
	    for (j = 0; j < a.num_columns(); ++j)
	    {
		cout.width(10);
		cout.precision(3);
		cout.setf(ios::right);
		cout.setf(ios::fixed, ios::floatfield);
		cout << a(i, j) << " ";
	    }
	    cout << endl;
	    ++i;
	}
    }
}

void print_s_trans(EST_Stream &a, int width)
{
    for (int i = 0; i < a.length(); ++i)
    {
	//	cout <<	(int)a.nth(i)->f("pos") << " XX " << a.nth(i)->name() << endl;
	if ((a.nth(i)->f("pos")==1) || (a.nth(i)->name() == "   "))
	{
	    //	    cout.setf(ios::fixed, ios::floatfield);
	    cout.width(width);
	    cout << a.nth(i)->name() << " ";
	}
    }
    cout << endl;
}

void make_hit_and_miss(EST_Stream &a)
{
    EST_Stream_Item *s;
    
    for (s = a.head(); s; s = next(s))
    {
	if (s->f("pos") == 0)
	    s->set_name(".");
	else if (s->f("hit") == 1)
	    s->set_name("HIT");
	else
	    s->set_name("MISS");
	s->f.clear();
    }
}

void pos_only(EST_Stream &a)
{
    EST_Stream_Item *a_ptr;
    
    for (a_ptr = a.head(); a_ptr; )
	if (a_ptr->f("pos")==0)
	{
	    a_ptr = a.remove(a_ptr);
	    if (a_ptr == 0)		// check to see if we are at head of list
		a_ptr = a.head();
	    else
		a_ptr = next(a_ptr);
	}
	else
	    a_ptr = next(a_ptr);
}

// Warning this is bugged - slight misalignments occur.
void print_aligned_trans(EST_Stream &ref, EST_Stream &test, EST_FMatrix &m)
{
    int i, j, n;
    EST_Stream al, refal, testal;
    EST_Stream_Item *p;
    EST_Stream_Item pos, blank;
    
    blank.set_name("   ");
    
    pos.f.set("pos", 1);
    blank.f.set("pos", 0);
    
    pos_only(test);
    pos_only(ref);

    // first check for emtpy matrices - indicates all insertions or deletions

    if ((m.num_columns() == 0) && (m.num_rows() != 0))
    {
	cout << "REC: ";
	print_s_trans(test);
	return;
    }
    else if ((m.num_columns() != 0) && (m.num_rows() == 0))
    {
	cout << "LAB: ";
	print_s_trans(ref);
	return;
    }
    else if ((m.num_columns() == 0) && (m.num_rows() == 0))
    {
	cout << "LAB: ";
	print_s_trans(ref);
	return;
    }

    int l;
    l = 0;
    
    //    cout << "ref: " << ref.name() << endl << ref;
    //    cout << "test: " << test.name() << endl << test;
    
    if (m(0, 0) < 0)
	refal.append(blank);
    
    
    pos.set_name(ref.head()->name());
    refal.append(pos);	
    for (i = 1, p = next(ref.head()); p; p = next(p), ++i)
    {
	n = num_b_insertions(m, l, i);
	
	for (j = 0; j < n; ++j)
	    refal.append(blank);
	
	if (n > -0.5)
	    l = i;
	pos.set_name(p->name());
	
	refal.append(pos);
    }
    
    l = 0;
    pos.set_name(test.head()->name());
    testal.append(pos);
    for (i = 1, p = next(test.head()); p; p = next(p), ++i)
    {
	n = num_b_deletions(m, l, i);
	
	//	cout << *p << "last " << l << " current " << i <<
	//	    " insertions " << n << endl;
	for (j = 0; j < n; ++j)
	    testal.append(blank);
	
	if (n > -0.5)
	    l = i;
	pos.set_name(p->name());
	testal.append(pos);
    }
    
    cout << "LAB: ";
    print_s_trans(refal, 3);
    cout << "REC: ";
    print_s_trans(testal, 3);
    
}

void print_results(EST_Stream &ref, EST_Stream &test, EST_FMatrix &m, int tot,
		   int del, int ins, int v)
{
    (void) tot;
    (void) del;
    (void) ins;
    if (v == 0)
	return;
    
    // v == 1 prints out total insertions etc

    if (v == 2)
    {
	cout << basename(ref.name(), "") << endl;
	print_i_d_scores(m);
	cout << endl;
    }
    else if (v == 3)
    {
	cout << basename(ref.name(), "") << endl;
	print_aligned_trans(ref, test, m);
	print_i_d_scores(m);
	cout << endl;
    }
    else if (v == 4)
    {
	cout << basename(ref.name(), "") << endl;
	print_matrix_scores(ref, test, m);
	print_i_d_scores(m);
	cout << endl;
    }
    else if (v == 5)
    {
	cout << basename(ref.name(), "") << endl;
	print_matrix_scores(ref, test, m);
	print_aligned_trans(ref, test, m);
	print_i_d_scores(m);
	cout << endl;
    }
    else if (v == 6)
    {
	print_matrix_scores(ref, test, m);
	error_location(ref, m, 1);
	make_hit_and_miss(ref);
	error_location(test, m, 0);
	make_hit_and_miss(test);
	ref.save("ref.error", "esps", "features");
	test.save("test.error", "esps", "features");
    }
    else if (v == 7)
    {
	error_location(ref, m, 1);
	make_hit_and_miss(ref);
	error_location(test, m, 0);
	make_hit_and_miss(test);
	ref.save("ref.error", "esps", "features");
	test.save("test.error", "esps", "features");
    }
}
